from typing import Optional

from sklearn import metrics
from transformers import BertConfig, Trainer, TrainingArguments
from transformers.trainer_utils import EvaluationStrategy

from .dataset import DatasetFactory
from .domain import Params
from .model import Classification
from .validation import summary


class Pipeline:
    def __init__(self, params: Params):
        self.init_dataset(params)
        self.init_model(params)

    def init_model(self, params):
        config = BertConfig()
        config.dropout_prob = params.dropout_prob
        config.num_labels = self.num_labels
        config.identifier = params.identifier
        self.config = config
        self.model = Classification(config)

    def init_dataset(self, params: Params):
        df = DatasetFactory(params)
        self.dataset_factory = df
        self.training_set, self.testing_set = df.supply_training_dataset()
        self.mlb = df.mlb
        self.tokenizer = df.tokenizer
        self.num_labels = df.num_labels

    def train(self, training_args: Optional[TrainingArguments] = None):
        if training_args is None:
            training_args = TrainingArguments(
                output_dir="./results",
                num_train_epochs=10,
                per_device_train_batch_size=8,
                save_steps=1000,
                save_total_limit=2,
                evaluation_strategy=EvaluationStrategy.EPOCH,
                logging_dir="./logs",
            )

        self.model.train()
        trainer = Trainer(
            model=self.model,
            args=training_args,
            train_dataset=self.training_set,
            eval_dataset=self.testing_set,
            compute_metrics=self._compute_metrics,
        )

        trainer.train()
        trainer.evaluate()
        self.trainer = trainer

    def _compute_metrics(self, pred):
        labels = pred.label_ids
        preds = pred.predictions >= 0
        precision, recall, f1, _ = metrics.precision_recall_fscore_support(
            labels, preds, average="micro"
        )
        acc = metrics.accuracy_score(labels, preds)
        return {
            "accuracy": acc,
            "f1": f1,
            "precision": precision,
            "recall": recall,
        }

    def validation_examples(self):
        assert self.trainer is not None, "training first"
        pred = self.trainer.predict(self.testing_set)
        cases = self.dataset_factory.x_test_dict
        true_tags = self.dataset_factory.y_test_tags
        pred_tags = self.dataset_factory.mlb.inverse_transform(pred.predictions >= 0)
        example, judges_count = summary(cases, true_tags, pred_tags)
        print(judges_count)
        return example, judges_count


if __name__ == "__main__":
    import fire

    print(fire.Fire(Params))



from .io_utils import load_state
from .web import Server
from .io_utils import prepare_state
import os

os.makedirs("data", exist_ok=True)
state_p = "data/unstate.pkl"
if not os.path.exists(state_p):
    prepare_state(state_p)
state = load_state(state_p)
server = Server(state)
server.plot()
server.app.run_server(debug=True)


from collections import Counter

import pandas as pd
import plotly.express as px
from sklearn import metrics
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from .data_utils import count_tags
from .domain import Mlb, RawData, States
from .model import StandaloneModel


def get_unlabelled_state(model: StandaloneModel, cases: list, mlb: Mlb):
    def tags_to_str(tags):
        return ", ".join(sorted(tags))

    k = len(cases)
    pooled_outputs = model.predict(cases, pooled_output=True)
    pred_tags = model.predict_tags(pooled_outputs, mlb)
    pred_tag_note = list(map(tags_to_str, pred_tags))
    index = list(range(k))
    tag_y = pred_tag_note
    tag_n = list(map(lambda tags: len(tags), pred_tags))
    from_ = ["unlabelled" for _ in range(k)]
    states = States(pooled_outputs, tag_y, index, tag_n, from_, pred_tag_note)
    return states


def get_tag_states(model: StandaloneModel, rawdata: RawData, mlb: Mlb):
    def tags_to_str(tags):
        return ", ".join(sorted(tags))

    x = rawdata.x_train_dict + rawdata.x_test_dict
    y = rawdata.y_train_tags + rawdata.y_test_tags
    index = list(range(len(rawdata.x_train_dict)))
    index.extend(range(len(rawdata.x_test_dict)))
    from_ = ["train" for _ in range(len(rawdata.x_train_dict))]
    from_.extend("test" for _ in range(len(rawdata.x_test_dict)))
    tag_n = list(map(lambda tags: len(tags), y))
    tag_y = list(map(tags_to_str, y))
    pooled_outputs = model.predict(x, pooled_output=True)
    pred_tags = model.predict_tags(x, mlb)
    pred_tag_note = list(map(tags_to_str, pred_tags))
    states = States(pooled_outputs, tag_y, index, tag_n, from_, pred_tag_note)
    return states


def dimension_reduction_plot(
    states: States, method_n="PCA", n_components=3, dash=False
):
    if method_n.lower() == "tsne":
        method = TSNE
    else:
        method = PCA
    dimension_reducer = method(n_components=n_components)
    result = dimension_reducer.fit_transform(states.data)
    if isinstance(dimension_reducer, PCA):
        print(
            f"Explained variation per principal component: {dimension_reducer.explained_variance_ratio_}"
        )
    df = pd.DataFrame(
        {
            "tag": states.tag,
            "index": states.index,
            "tag_num": states.tag_n,
            "from": states.from_,
            "pred_tag": states.pred_tag,
        }
    )
    for n in range(n_components):
        df[f"D{n+1}"] = result[:, n]

    if n_components == 3:
        fig = px.scatter_3d(
            df,
            x="D1",
            y="D2",
            z="D3",
            color="tag",
            symbol="tag_num",
            hover_data=["index", "from", "pred_tag"],
        )
    elif n_components == 2:
        fig = px.scatter(
            df,
            x="D1",
            y="D2",
            color="tag",
            symbol="tag_num",
            hover_data=["index", "from", "pred_tag"],
        )
    else:
        print("support only 2 or 3 dimension ploting")
        return
    fig.layout.update(showlegend=False)
    if dash:
        return fig, dimension_reducer
    fig.show()


def judge_on_tag(model: StandaloneModel, mlb: Mlb, rawdata: RawData):
    x = rawdata.x_test_dict
    y = rawdata.y_test_tags
    total_y = rawdata.y_tags
    pred_prob = model.predict(x)
    preds = pred_prob >= 0.5
    mcm = metrics.multilabel_confusion_matrix(mlb.transform(y), preds)
    ability = list(map(compress, mcm))
    tag_count = count_tags(total_y)
    sample_sizes = [tag_count[class_] for class_ in mlb.classes_]
    performance = pd.DataFrame(
        {
            "Tag": mlb.classes_,
            "Acc": [pair[0] for pair in ability],
            "Num": [pair[2] for pair in ability],
            "Sample_Size": sample_sizes,
        }
    )
    fig = px.scatter(performance, x="Tag", y="Acc", size="Sample_Size", color="Num")
    fig.show()


def compress(cm):
    err, corr = cm[1]
    amount = err + corr
    if amount == 0:
        return (0, corr, amount)
    return (corr / amount, corr, amount)


def summary(cases, true_tags, pred_tags):
    example = []
    judges = []
    for case, pred_tag, true_tag in zip(cases, pred_tags, true_tags):
        num_tags = len(pred_tag)
        corr = sum(tag in true_tag for tag in pred_tag)
        if num_tags == 0:
            judge = "missing"
        elif corr == num_tags:
            judge = "correct"
        else:
            judge = f"{corr} in {num_tags} tags correct"
        example.append((case, "; ".join(pred_tag), "; ".join(true_tag), judge))
        judges.append(judge)
    return example, Counter(judges)



import random
from typing import Dict, List

from . import data_utils, io_utils
from .domain import LabelledCase, Labels, Mlb
from .model import StandaloneModel

random.seed(42)


def edit_review(old_cases: List[LabelledCase], reviews: List[LabelledCase]):
    cases_copy = old_cases.copy()
    review_map = {"".join(review.text.values()): review.tag for review in reviews}
    used_keys = set()
    for idx, case in enumerate(old_cases):
        key = "".join(case.text.values())
        cases_copy[idx].tag = data_utils.add_acute_LL(cases_copy[idx].tag)
        if key in review_map:
            cases_copy[idx].tag = review_map[key]
            used_keys.add(key)

    for review_idx, key in enumerate(review_map):
        if key not in used_keys:
            cases_copy.append(reviews[review_idx])
    return cases_copy


def review_pipe(datazip_path: str, review_path: str):
    raw_data = io_utils.load_datazip(datazip_path)
    x = raw_data.x_dict
    y = raw_data.y_tags
    labelled_cases = [LabelledCase(k, v) for k, v in zip(x, y)]
    reviews = data_utils.load_labelled_cases(review_path)
    reviewed = edit_review(labelled_cases, reviews)
    new_x, new_y = data_utils.labelled_cases_to_xy(reviewed)
    zipname = data_utils.split_and_dump_dataset(new_x, new_y)
    return zipname


def enrich(
    model: StandaloneModel,
    mlb: Mlb,
    all_cases,
    labelled_cases: List[LabelledCase],
    thresh=20,
):
    known_cases, known_tags = data_utils.unwrap_labelled_cases(labelled_cases)
    unlabelled_cases = data_utils.cases_minus(all_cases, known_cases)
    pred_tags = model.predict_tags(unlabelled_cases, mlb)
    needed = get_needed(known_tags, pred_tags, thresh=thresh)
    collection = collect(unlabelled_cases, needed, pred_tags)
    io_utils.dump_labelled_cases(
        collection, f"enrich_{data_utils.get_timestamp()}.json"
    )


def get_needed(known_tags: Labels, pred_tags: Labels, thresh=20):
    def sampleable(tag, lib: Dict[str, list], need: dict):
        return len(lib.get(tag, [])) > need[tag]

    have = data_utils.count_tags(known_tags)
    need_num = {}
    for tag, num in have.items():
        if num < thresh:
            need_num[tag] = thresh - num
    lib = data_utils.grouping_idx(pred_tags)

    needed = {
        tag: random.sample(lib[tag], need_num[tag])
        for tag in need_num
        if sampleable(tag, lib, need_num)
    }
    return needed


def collect(unlabelled_cases, needed: dict, pred_tags: list):
    collection: List[LabelledCase] = []
    for indexes in needed.values():
        collection.extend(
            LabelledCase(unlabelled_cases[idx], pred_tags[idx]) for idx in indexes
        )
    return collection
